﻿module Fluid

// This module implements the Lattice Bolztmann Method (LBM) for fluid simulation

[<Struct>]
type lbm_data_t =
    val weight : float32;
    val dxi : int;
    val dyi : int;
    val dxf : float32;
    val dyf : float32;
    val opposite_idx : int;
    val free_slip_idx1 : int;
    val free_slip_idx2 : int;
    
    new (w, dx, dy, opp_idx, fs1, fs2) = { 
        weight = w; 
        dxi = dx; 
        dyi = dy; 
        dxf = float32 dx; 
        dyf = float32 dy; 
        opposite_idx = opp_idx; 
        free_slip_idx1 = fs1;
        free_slip_idx2 = fs2;
    }

type field = {
    field_data_cur : float32 array;
    field_data_tmp : float32 array;
    field_blocked : bool array;
    field_w : int;
    field_h : int;
}

let lbm_data = [|
    new lbm_data_t(4.f /  9.f,  0,  0,  0,  -1, -1); // 0
    new lbm_data_t(1.f /  9.f,  1,  0,  3,  -1, -1); // 1
    new lbm_data_t(1.f /  9.f,  0,  1,  4,  -1, -1); // 2
    new lbm_data_t(1.f /  9.f, -1,  0,  1,  -1, -1); // 3
    new lbm_data_t(1.f /  9.f,  0, -1,  2,  -1, -1); // 4
    new lbm_data_t(1.f / 36.f,  1,  1,  7,   6,  8); // 5
    new lbm_data_t(1.f / 36.f, -1,  1,  8,   5,  7); // 6
    new lbm_data_t(1.f / 36.f, -1, -1,  5,   8,  6); // 7
    new lbm_data_t(1.f / 36.f,  1, -1,  6,   7,  5); // 8
|]

let smoothness = 0.59f
let smoothness_one_minus = 1.f - smoothness

let dampening_mul = 0.99f

let create w h =
    let amt = w * h * 9
    let fld =
        { field_data_cur = Array.create amt 0.f;
          field_data_tmp = Array.create amt 0.f;
          field_blocked = Array.create (w * h) false;
          field_w = w;
          field_h = h }
          
    // Set the initial velocities
    (*for y = 0 to h - 1 do
        for x = 0 to w - 1 do
            let idx9 = (x + y * w) * 9
          
            let dens = 0.f
            let vel_x = 0.1f
            let vel_y = 0.f
            
            // Calculate the equilibrium distribution
            let vel_sq = vel_x * vel_x + vel_y * vel_y
            let vel_sq_3_2 = (3.f / 2.f) * vel_sq
            for i = 0 to 8 do
                let dx = lbm_data.[i].dxf
                let dy = lbm_data.[i].dyf
                let dot = dx * vel_x + dy * vel_y
                let f_eq = 
                    lbm_data.[i].weight *
                    (dens + 3.f * dot - vel_sq_3_2 + (9.f / 2.f) * dot * dot)
                    
                fld.field_data_cur.[idx9 + i] <- smoothness * f_eq*)
            
    fld

let get_w fld = fld.field_w
let get_h fld = fld.field_h

let get_density fld x y =
    if x >= 0 && y >= 0 && x < fld.field_w && y < fld.field_h then
        let src = fld.field_data_cur
        let idx9 = (x + y * fld.field_w) * 9
        let mutable dens = 0.f
        for i = 0 to 8 do
            let v = src.[idx9 + i]
            dens <- dens + v
        dens
    else
        0.f

let get_velocity fld x y =
    if x >= 0 && y >= 0 && x < fld.field_w && y < fld.field_h then
        let src = fld.field_data_cur
        let idx9 = (x + y * fld.field_w) * 9
        let mutable vel_x = 0.f
        let mutable vel_y = 0.f
        for i = 0 to 8 do
            let v = src.[idx9 + i]
            vel_x <- vel_x + lbm_data.[i].dxf * v
            vel_y <- vel_y + lbm_data.[i].dyf * v
        (vel_x, vel_y)
    else
        (0.f, 0.f)

let get_velocity_smooth fld (xf:float32) (yf:float32) =
    let xs = floor xf
    let ys = floor yf
    let dx = xf - xs
    let dy = yf - ys
    let dx1 = 1.f - dx
    let dy1 = 1.f - dy
    let x = int xs
    let y = int ys
    
    // Get the velocities
    let vx1, vy1 = get_velocity fld (x + 0) (y + 0)
    let vx2, vy2 = get_velocity fld (x + 1) (y + 0)
    let vx3, vy3 = get_velocity fld (x + 0) (y + 1)
    let vx4, vy4 = get_velocity fld (x + 1) (y + 1)
    
    // Interpolate
    let vx = (vx1 * dx + vx2 * dx1) * dy + (vx3 * dx + vx2 * dx1) * dy1
    let vy = (vy1 * dx + vy2 * dx1) * dy + (vy3 * dx + vy2 * dx1) * dy1
    (vx, vy)

let get_blocked fld x y =
    if x >= 0 && y >= 0 && x < fld.field_w && y < fld.field_h then
        fld.field_blocked.[x + y * fld.field_w]
    else
        true

let set_blocked fld x y is_blocked =
    if x >= 0 && y >= 0 && x < fld.field_w && y < fld.field_h then
        fld.field_blocked.[x + y * fld.field_w] <- is_blocked

let update fld =
    let src = fld.field_data_cur
    let dest = fld.field_data_tmp
    let blocked = fld.field_blocked
    let w = fld.field_w
    let h = fld.field_h
    
    let clear_column = w - 1
    
    // Tests if a cell is blocked
    let is_blocked x y =
        if x < 0 || y < 0 || x >= w || y >= h then
            true
        else
            blocked.[x + y * w]

    // Stream and collide (no-slip)
    for y = 0 to h - 1 do
        for x = 0 to w - 1 do
            let idx9 = (x + y * w) * 9
            
            // If the cell is blocked, then clear it
            if (is_blocked x y) || (x >= clear_column) then
                for i = 0 to 8 do
                    dest.[idx9 + i] <- 0.f
            else
                // Stream from adjacent cells
                for i = 0 to 8 do
                    let dx = -lbm_data.[i].dxi
                    let dy = -lbm_data.[i].dyi
                    
                    // Check for boundary condition
                    if is_blocked (x + dx) (y + dy) then
                        // No-slip boundary condition
                        dest.[idx9 + i] <- src.[idx9 + lbm_data.[i].opposite_idx]
                    else
                        dest.[idx9 + i] <- src.[idx9 + ((dx + dy * w) * 9) + i]
                      
    // Stream and collide (free-slip)
    (*for i = 0 to dest.Length - 1 do
        dest.[i] <- 0.f

    for y = 0 to h - 1 do
        for x = 0 to w - 1 do
            let idx9 = (x + y * w) * 9
            
            // If the cell is blocked, then clear it
            if (is_blocked x y) || (x >= clear_column) then
                ()
            else
                // Stream to adjacent cells
                for i = 0 to 8 do
                    let dx = lbm_data.[i].dxi
                    let dy = lbm_data.[i].dyi
                    
                    let v = src.[idx9 + i]
                    
                    // Check for boundary condition
                    if is_blocked (x + dx) (y + dy) then
                        if i <= 4 then
                            // Bounce back to the same cell
                            let opp_i = lbm_data.[i].opposite_idx
                            dest.[idx9 + opp_i] <- dest.[idx9 + opp_i] + v
                        else
                            // Try the two free slip directions
                            let can_slip1 = not (is_blocked x (y + dy))
                            let can_slip2 = not (is_blocked (x + dx) y)
                            let slip1_idx = idx9 + dy * w * 9 + lbm_data.[i].free_slip_idx1
                            let slip2_idx = idx9 + dx     * 9 + lbm_data.[i].free_slip_idx2
                            
                            if (can_slip1 && can_slip2) then
                                dest.[slip1_idx] <- dest.[slip1_idx] + v * 0.5f
                                dest.[slip2_idx] <- dest.[slip2_idx] + v * 0.5f
                            elif can_slip1 then
                                dest.[slip1_idx] <- dest.[slip1_idx] + v
                            elif can_slip2 then
                                dest.[slip2_idx] <- dest.[slip2_idx] + v
                            else
                                let opp_i = lbm_data.[i].opposite_idx
                                dest.[idx9 + opp_i] <- dest.[idx9 + opp_i] + v
                    else
                        // Normal streaming into adjacent cell
                        let dadd = (dx + dy * w) * 9
                        dest.[idx9 + dadd + i] <- dest.[idx9 + dadd + i] + v
                   
    for y = 0 to h - 1 do
        for x = 0 to w - 1 do
            // If the cell is blocked, then clear it
            if (is_blocked x y) || (x >= clear_column) then
                let idx9 = (x + y * w) * 9
                for i = 0 to 8 do
                    dest.[idx9 + i] <- 0.f*)
                        
    // Update the distribution functions at each cell (calculate the equilibrium state)
    for y = 0 to h - 1 do
        for x = 0 to w - 1 do
            let idx9 = (x + y * w) * 9
            
            if (is_blocked x y) || (x >= clear_column) then
                ()
            else
                // Calculate the density and velocity
                let mutable dens = 0.f
                let mutable vel_x = 0.f
                let mutable vel_y = 0.f

                for i = 0 to 8 do
                    let v = dest.[idx9 + i]
                    dens <- dens + v
                    vel_x <- vel_x + lbm_data.[i].dxf * v
                    vel_y <- vel_y + lbm_data.[i].dyf * v
                  
                if x = 0 then vel_x <- 0.1f
                    
                // Calculate the equilibrium distribution
                let vel_sq = vel_x * vel_x + vel_y * vel_y
                let vel_sq_3_2 = (3.f / 2.f) * vel_sq
                for i = 0 to 8 do
                    let dx = lbm_data.[i].dxf
                    let dy = lbm_data.[i].dyf
                    let dot = dx * vel_x + dy * vel_y
                    let f_eq = 
                        lbm_data.[i].weight *
                        (dens + 3.f * dot - vel_sq_3_2 + (9.f / 2.f) * dot * dot)
                        
                    dest.[idx9 + i] <- (smoothness_one_minus * dest.[idx9 + i] + smoothness * f_eq) * dampening_mul
    
    { fld with field_data_cur = dest; field_data_tmp = src }
