﻿module Islands
open System.Drawing
open System.Drawing.Imaging

type field = {
    field_w : int;
    field_h : int;
    field_data : float32 array;
    field_bitmap : Bitmap;
}

let color_from_value v =
    let alpha = 
        if v < 0.4f then 
            Util.map_range 0.f 0.4f 0.f 0.1f v
        else
            let a = Util.map_range 0.4f 1.f 0.1f 1.f v
            Util.pow a 3.f
            
    let r = Util.map_range 0.f 1.f 0.7f 0.96f v
    let g = Util.map_range 0.f 1.f 0.4f 0.7f v
    let b = Util.map_range 0.f 1.f 0.1f 0.2f v
    
    Color.FromArgb(Util.color_f2i alpha, Util.color_f2i r, Util.color_f2i g, Util.color_f2i b)

let get_bitmap fld =
    fld.field_bitmap

let get fld x y =
    if x >= 0 && y >= 0 && x < (fld.field_w) && (y < fld.field_h) then
        fld.field_data.[x + y * fld.field_w]
    else
        0.f

let get_clamp fld x y =
    let x = Util.clamp x 0 (fld.field_w - 1)
    let y = Util.clamp y 0 (fld.field_h - 1)
    fld.field_data.[x + y * fld.field_w]

let set fld x y v =
    if x >= 0 && y >= 0 && x < (fld.field_w) && (y < fld.field_h) then
        let new_v = Util.clamp v 0.f 1.f
        fld.field_data.[x + y * fld.field_w] <- new_v
        fld.field_bitmap.SetPixel(x, y, color_from_value new_v)

let add fld x y v =
    if x >= 0 && y >= 0 && x < (fld.field_w) && (y < fld.field_h) then
        let idx = x + y * fld.field_w
        let new_v = Util.clamp (fld.field_data.[idx] + v) 0.f 1.f
        fld.field_data.[idx] <- new_v
        fld.field_bitmap.SetPixel(x, y, color_from_value new_v)

let random_add fld x y amt size v =
    for i = 1 to amt do
        let ang = Util.rand_float 0.f ((float32 System.Math.PI) * 2.f)
        let dist = (Util.rand_float 0.f size) * (Util.rand_float 0.f 1.f)
        let xd = x + int ((cos ang) * dist)
        let yd = y + int ((sin ang) * dist)
        add fld xd yd v

let is_full fld x y xs ys =
    let mutable found = 0
    for ya = 0 to ys - 1 do
        for xa = 0 to xs - 1 do
            if get_clamp fld (x + xa) (y + ya) >= 0.7f then
                found <- found + 1
    found > ((xs * ys) * 60 / 100)

let clear fld =
    let w = fld.field_w
    let h = fld.field_h
    for y = 0 to h - 1 do
        for x = 0 to w - 1 do
            set fld x y (Util.rand_float 0.f 0.2f)

let create w h =
    let size = w * h
        
    let fld =
        { field_w = w;
          field_h = h;
          field_data = Array.create size 0.f;
          field_bitmap = new Bitmap(w, h, PixelFormat.Format32bppArgb) }

    clear fld
            
    fld
